"""Tests for Heroku migrate script."""

from mock import call
from pyramid_heroku.migrate import alembic
from pyramid_heroku.migrate import migrate
from pyramid_heroku.migrate import needs_migrate

import mock
import pytest
import responses
import unittest


class TestHerokuMigrate(unittest.TestCase):
    def setUp(self):
        from pyramid_heroku.heroku import Heroku

        self.Heroku = Heroku

    def test_default_values(self):
        from pyramid_heroku.migrate import main

        def assert_args(args, called_with):
            with mock.patch("sys.argv", ["migrate.py"] + args), mock.patch(
                "pyramid_heroku.migrate.Heroku"
            ) as heroku, mock.patch("pyramid_heroku.migrate.migrate"):
                main()
            heroku.assert_called_with(*called_with)

        # all default
        assert_args(["my_test_app"], ["my_test_app", "etc/production.ini"])

        # all custom
        assert_args(
            ["my_test_app", "etc/develop2.ini"], ["my_test_app", "etc/develop2.ini"]
        )

    @mock.patch("pyramid_heroku.subprocess")
    @responses.activate
    def test_needs_migrate(self, sub):
        h = self.Heroku("test", "etc/production.ini")
        sub.run.stdout.return_value = b""
        self.assertTrue(needs_migrate(h))
        sub.run.assert_called_with(
            [
                "alembic",
                "-c",
                "etc/alembic.ini",
                "-x",
                "ini=etc/production.ini",
                "current",
            ],
            stdout=mock.ANY,
            stderr=mock.ANY,
        )

    @mock.patch("pyramid_heroku.subprocess")
    @responses.activate
    def test_does_not_need_migrate(self, sub):
        h = self.Heroku("test", "etc/production.ini")
        p = mock.Mock()
        p.stdout = b"head"
        sub.run.return_value = p
        self.assertFalse(needs_migrate(h))
        sub.run.assert_called_with(
            [
                "alembic",
                "-c",
                "etc/alembic.ini",
                "-x",
                "ini=etc/production.ini",
                "current",
            ],
            stdout=mock.ANY,
            stderr=mock.ANY,
        )

    @mock.patch("pyramid_heroku.subprocess")
    @responses.activate
    def test_alembic(self, sub):
        h = self.Heroku("test", "etc/production.ini")
        p = mock.Mock()
        p.stdout = b"head"
        sub.run.return_value = p
        alembic(h)
        # TODO: print(migration done)
        sub.run.assert_called_with(
            [
                "alembic",
                "-c",
                "etc/alembic.ini",
                "-x",
                "ini=etc/production.ini",
                "upgrade",
                "head",
            ],
            stdout=mock.ANY,
            stderr=mock.ANY,
        )

    @mock.patch("pyramid_heroku.subprocess")
    @mock.patch("pyramid_heroku.migrate.print")
    @mock.patch("pyramid_heroku.migrate.sleep")
    @responses.activate
    def test_migrate_skip(self, sleep, out, sub):
        h = self.Heroku("test", "etc/production.ini")
        p = mock.Mock()
        p.stdout = b"head"
        sub.run.return_value = p
        migrate(h)
        sub.run.assert_called_with(
            [
                "alembic",
                "-c",
                "etc/alembic.ini",
                "-x",
                "ini=etc/production.ini",
                "current",
            ],
            stdout=mock.ANY,
            stderr=mock.ANY,
        )
        out.assert_has_calls([call("Database migration is not needed")], any_order=True)

    @mock.patch("pyramid_heroku.subprocess")
    @mock.patch("pyramid_heroku.migrate.print")
    @mock.patch("pyramid_heroku.migrate.sleep")
    @mock.patch("pyramid_heroku.heroku.Session")
    @responses.activate
    def test_migrate_with_maintenance_mode(self, ses, sleep, out, sub):

        h = self.Heroku("test", "etc/production.ini")
        h.get_maintenance = mock.Mock(return_value=False)
        h.set_maintenance = mock.Mock()
        migrate(h)

        h.set_maintenance.assert_has_calls([call(True), call(False)])
        sub.run.assert_called_with(
            [
                "alembic",
                "-c",
                "etc/alembic.ini",
                "-x",
                "ini=etc/production.ini",
                "upgrade",
                "head",
            ],
            stdout=mock.ANY,
            stderr=mock.ANY,
        )

    @mock.patch("pyramid_heroku.subprocess")
    @mock.patch("pyramid_heroku.migrate.print")
    @mock.patch("pyramid_heroku.migrate.sleep")
    @mock.patch("pyramid_heroku.heroku.Session")
    @responses.activate
    def test_migrate_skip_setting_maintenance_mode(self, ses, sleep, out, sub):

        h = self.Heroku("test", "etc/production.ini")
        h.get_maintenance = mock.Mock(return_value=True)
        h.set_maintenance = mock.Mock()
        migrate(h)

        assert not h.set_maintenance.called
        sub.run.assert_called_with(
            [
                "alembic",
                "-c",
                "etc/alembic.ini",
                "-x",
                "ini=etc/production.ini",
                "upgrade",
                "head",
            ],
            stdout=mock.ANY,
            stderr=mock.ANY,
        )

    @mock.patch("pyramid_heroku.subprocess")
    @mock.patch("pyramid_heroku.print")
    @mock.patch("pyramid_heroku.migrate.sleep")
    @mock.patch("pyramid_heroku.heroku.Session")
    @responses.activate
    def test_migrate_non_zero(self, ses, sleep, out, sub):
        h = self.Heroku("test", "etc/production.ini")
        p = mock.Mock()
        p.stdout = b"foo"
        p.stderr = b"bar"
        p.check_returncode.side_effect = Exception
        sub.run.return_value = p

        with pytest.raises(Exception):
            migrate(h)

        out.assert_has_calls([call("foo"), call("bar", file=mock.ANY)], any_order=True)



"""Set client_addr IP that we can trust."""
from ipaddress import ip_address
from ipaddress import ip_network

import logging
import requests


def includeme(config):
    config.add_tween("pyramid_heroku.client_addr.ClientAddr")


class ClientAddr(object):
    """Set request.client_addr to an IP we can trust.

    Sets correct IP to request.client_addr when running on Heroku with
    gunicorn. Should execute before other tweens which depend on it. This can
    be achieved with :ref:`Explicit tween ordering
    <pyramid:explicit_tween_ordering>` or by passing 'over' and 'under'
    arguments to :py:meth:`config.Configurator.add_tween` like this

    .. code-block:: python

        config.add_tween(
            'pyramid_heroku.client_addr.ClientAddr',
            under=pyramid.tweens.INGRESS,
        )

    The tween checks if ``X-Forwarded-For`` is present and if it is, it
    filters out any Cloudflare IPs and takes the last IP in the
    ``X-Forwarded-For`` list and makes it the only IP in the list.

    Cloudflare IPs are filtered out to avoid setting Cloudflare IP in
    ``X-Forwarded-For`` when using Cloudflare's reverse proxy.

    This effectively causes pyramid to return this real IP for
    ``request.client_addr``. Read rationale behind why we do this on
    https://github.com/teamniteo/heroku_ips.
    """

    def __init__(self, handler, registry):
        self.handler = handler
        self.registry = registry

        if self.registry.settings.get("pyramid_heroku.structlog"):
            import structlog

            self.logger = structlog.getLogger(__name__)
        else:
            self.logger = logging.getLogger(__name__)

        self.ignored_ip_networks = self.get_cloudflare_ip_networks()

    def __call__(self, request):
        if request.headers.get("X-Forwarded-For"):
            ips = [ip.strip() for ip in request.headers["X-Forwarded-For"].split(",")]
            filtered_ips = list(self.remove_ignored_ips(ips))

            request.headers["X-Forwarded-For"] = (
                filtered_ips[-1] if filtered_ips else ""
            )
        return self.handler(request)

    def remove_ignored_ips(self, ips):
        """Remove ignored IPs from the list."""
        for ip in ips:
            try:
                address = ip_address(ip)
            except ValueError as e:
                self.logger.info("Ignoring invalid IP address", exc_info=e)
                continue

            is_ignored = any(
                [address in network for network in self.ignored_ip_networks]
            )
            if not is_ignored:
                yield ip

    def get_ip_networks_from_url(self, url):
        """Get the list of IP networks from given url."""
        try:
            res = requests.get(url, timeout=10)
            res.raise_for_status()
            return [ip_network(n) for n in res.text.split()]
        except Exception as e:
            self.logger.exception(
                f"Failed getting a list of IPs from {url}", exc_info=e
            )

            return []

    def get_cloudflare_ip_networks(self):
        """Get the list of Cloudflare's current IP ranges."""
        ipv4 = self.get_ip_networks_from_url("https://www.cloudflare.com/ips-v4")
        ipv6 = self.get_ip_networks_from_url("https://www.cloudflare.com/ips-v6")
        return ipv4 + ipv6



"""Enable or disable Heroku app maintenance mode."""

from enum import Enum
from pyramid_heroku.heroku import Heroku

import argparse


class Mode(Enum):
    """Heroku maintenance modes."""

    on = "ON"
    off = "OFF"


def main() -> None:
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        usage=(
            "usage: maintenance.py [-h] [app_name] [ini_file] [on|off]"
            "\nexample: python -m pyramid_heroku.maintenance on my_app etc/production.ini"
        ),
    )
    parser.add_argument("mode", choices=[x.name for x in Mode], help="Maintenance mode")
    parser.add_argument("app_name", help="Heroku app name")
    parser.add_argument(
        "ini_file",
        nargs="?",
        default="etc/production.ini",
        help="Path to Pyramid configuration file ",
    )

    options = parser.parse_args()

    h = Heroku(options.app_name, options.ini_file)

    if Mode[options.mode] == Mode.on:
        h.set_maintenance(True)
    else:
        h.set_maintenance(False)


if __name__ == "__main__":
    main()



"""Tests for HerokuappAccess tween."""

from pyramid import testing
from zope.testing.loggingsupport import InstalledHandler

import logging
import mock
import structlog
import unittest

tweens_handler = InstalledHandler("pyramid_heroku.herokuapp_access")


class TestHerokuappAccessTween(unittest.TestCase):
    def wrap_logger(self, _, __, event_dict):
        """Have structlog output to the standard logger, so that tweens_handler can intercept its messages."""
        wrapped_logger = logging.getLogger("pyramid_heroku.herokuapp_access")
        wrapped_logger.info(event_dict["event"])
        return event_dict["event"]

    def setUp(self):
        tweens_handler.clear()
        self.config = testing.setUp()
        self.handler = mock.Mock()
        self.request = testing.DummyRequest()
        self.request.registry.settings = {
            "pyramid_heroku.herokuapp_whitelist": ["1.2.3.4"],
            "pyramid_heroku.structlog": 1,
        }
        structlog.configure(processors=[self.wrap_logger], context_class=dict)

    def tearDown(self):
        tweens_handler.clear()
        testing.tearDown()

    def test_whitelisted_ip(self):
        from pyramid_heroku.herokuapp_access import HerokuappAccess

        self.request.client_addr = "1.2.3.4"
        self.request.headers = {"Host": "foo.herokuapp.com"}

        HerokuappAccess(self.handler, self.request.registry)(self.request)
        self.handler.assert_called_with(self.request)

    def test_non_whitelisted_ip(self):
        from pyramid_heroku.herokuapp_access import HerokuappAccess

        self.request.client_addr = "6.6.6.6"
        self.request.headers = {"Host": "foo.herokuapp.com"}

        # structlog version
        # wrap structlog around a regular logger so that zope's logging support
        # still gets the logged messages
        response = HerokuappAccess(self.handler, self.request.registry)(self.request)
        assert not self.handler.called, "handler should not be called"
        self.assertEqual(len(tweens_handler.records), 1)
        self.assertEqual(
            "Herokuapp access denied", tweens_handler.records[0].msg  # noqa
        )
        self.assertEqual(response.status_code, 403)

        # standard logging version
        self.request.registry.settings["pyramid_heroku.structlog"] = False
        tweens_handler.clear()
        response = HerokuappAccess(self.handler, self.request.registry)(self.request)
        assert not self.handler.called, "handler should not be called"
        self.assertEqual(len(tweens_handler.records), 1)
        self.assertEqual(
            "Denied Herokuapp access for Host foo.herokuapp.com and IP 6.6.6.6",  # noqa
            tweens_handler.records[0].msg,
        )
        self.assertEqual(response.status_code, 403)

    def test_other_hostname(self):
        from pyramid_heroku.herokuapp_access import HerokuappAccess

        self.request.client_addr = "6.6.6.6"
        self.request.headers = {"Host": "foo.bar.com"}

        HerokuappAccess(self.handler, self.request.registry)(self.request)
        self.handler.assert_called_with(self.request)

    def test_herokuapp_whitelist_not_set(self):
        "Even if whitelist is not set, the protection should still work."
        from pyramid_heroku.herokuapp_access import HerokuappAccess

        self.request.client_addr = "6.6.6.6"
        self.request.headers = {"Host": "foo.herokuapp.com"}
        self.request.registry.settings = {}

        HerokuappAccess(self.handler, self.request.registry)(self.request)
        assert not self.handler.called, "handler should not be called"

    def test_herokuapp_whitelist_empty(self):
        "Even if whitelist is empty, the protection should still work."
        from pyramid_heroku.herokuapp_access import HerokuappAccess

        self.request.client_addr = "6.6.6.6"
        self.request.headers = {"Host": "foo.herokuapp.com"}
        self.request.registry.settings = {"pyramid_heroku.herokuapp_whitelist": []}

        HerokuappAccess(self.handler, self.request.registry)(self.request)
        assert not self.handler.called, "handler should not be called"



"""Tests for Host tween."""

from pyramid import request
from pyramid import testing

import mock
import unittest


class TestHostTween(unittest.TestCase):
    def setUp(self):
        self.config = testing.setUp()
        self.request = request.Request({})

        self.handler = mock.Mock()
        self.registry = None

    def tearDown(self):
        testing.tearDown()

    def test_direct_access(self):
        from pyramid_heroku.host import Host

        self.request.environ["HTTP_HOST"] = "foo.com"
        Host(self.handler, self.registry)(self.request)
        self.handler.assert_called_with(self.request)
        self.assertEqual(self.request.host, "foo.com")

    def test_proxy_access(self):
        from pyramid_heroku.host import Host

        self.request.environ["HTTP_HOST"] = "foo.com"
        self.request.headers["X-Forwarded-Host"] = "bar.com"

        Host(self.handler, self.registry)(self.request)
        self.handler.assert_called_with(self.request)
        self.assertEqual(self.request.host, "bar.com")



"""Prevent .herokuapp.com access to non-whitelisted IPs"""

from pyramid.response import Response

import logging


def includeme(config):
    config.add_tween("pyramid_heroku.herokuapp_access.HerokuappAccess")


class HerokuappAccess(object):
    """Deny access to 'herokuapp.com' urls for IPs that are not whitelisted.

    Denies access to hosts that contain 'herokuapp.com' to ips that are not
    whitelisted. Registry settings must contain a key
    'pyramid_heroku.herokuapp_whitelist', whose value must be a list of
    strings (IPs). For this tween to work, you must include the
    ClientAddr tween and make sure this tween executes after the ClientAddr
    tween.
    """

    def __init__(self, handler, registry):
        self.handler = handler
        self.registry = registry
        self.settings = getattr(registry, "settings", {})

    def __call__(self, request):
        whitelisted_ips = request.registry.settings.get(
            "pyramid_heroku.herokuapp_whitelist", []
        )

        if (
            "herokuapp.com" in request.headers["Host"]
            and request.client_addr not in whitelisted_ips
        ):
            if request.registry.settings.get("pyramid_heroku.structlog"):
                import structlog

                logger = structlog.getLogger(__name__)
                logger.info(
                    "Herokuapp access denied",
                    user_ip=request.client_addr,
                    host=request.headers["Host"],
                )
            else:
                logger = logging.getLogger(__name__)
                logger.info(
                    f'Denied Herokuapp access for Host {request.headers["Host"]}'
                    f" and IP {request.client_addr}"
                )

            resp = Response("Unauthorized.", status=403, content_type="text/plain")
            resp.status = "403 Unauthorized"
            return resp

        return self.handler(request)



"""Graceful DB migrations."""


from pyramid_heroku import shell
from pyramid_heroku.heroku import Heroku
from time import sleep

import argparse


def needs_migrate(heroku: Heroku) -> bool:
    """
    Checks current status of alembic branches.

    :return: True if current alembic branch is not up to date.
    """
    cmd = shell(f"alembic -c etc/alembic.ini -x ini={heroku.ini_file} current")
    return "head" not in cmd


def alembic(heroku: Heroku) -> None:
    """
    Run alembic migrations.
    """
    shell(f"alembic -c etc/alembic.ini -x ini={heroku.ini_file} upgrade head")


def migrate(heroku: Heroku) -> None:
    """Run database migration if needed."""
    print(heroku.app_name)
    print(heroku.ini_file)

    if needs_migrate(heroku):
        was_in_maintenance = heroku.get_maintenance()
        if not was_in_maintenance:
            heroku.set_maintenance(True)
        heroku.scale_down()
        sleep(30)
        alembic(heroku)
        heroku.scale_up()
        sleep(30)
        if not was_in_maintenance:
            # Don't disable maintenance mode if it was enabled externally.
            heroku.set_maintenance(False)
    else:
        print("Database migration is not needed")


def main() -> None:
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        usage=(
            "usage: migrate.py [-h] [app_name] [ini_file]"
            "\nexample: python -m pyramid_heroku.migrate my_app etc/production.ini"
        ),
    )
    parser.add_argument("app_name", help="Heroku app name")
    parser.add_argument(
        "ini_file",
        nargs="?",
        default="etc/production.ini",
        help="Path to Pyramid configuration file ",
    )

    options = parser.parse_args()

    migrate(Heroku(options.app_name, options.ini_file))


if __name__ == "__main__":
    main()



"""Various utilities."""

from ast import literal_eval
from expandvars import expandvars

import shlex
import subprocess
import sys
import typing as t


def shell(cmd: str) -> str:
    """
    Run shell command.

    :param cmd: shell command to run
    :return: stdout of command
    """

    p = subprocess.run(shlex.split(cmd), stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    print(p.stdout.decode())
    print(p.stderr.decode(), file=sys.stderr)
    p.check_returncode()
    return p.stdout.decode()


def safe_eval(text: str) -> t.Optional[str]:
    """Safely evaluate `text` argument.

    `text` can be evaluated to string, number,
    tuple, list, dict, boolean, and None.
    Code `text[0].upper + text[1:]` is for lower case
    'false' or 'true' strings, it should not break anything.
    """
    if not isinstance(text, str):
        raise ValueError(f"Expected a string, got {type(text)} instead.")

    if len(text) == 0:
        return text

    try:
        return literal_eval(text[0].upper() + text[1:])
    except (ValueError, SyntaxError):
        return text


def expandvars_dict(settings: t.Dict[str, str]) -> t.Dict[str, t.Any]:
    """Expand strings or variables from the environment into equivalent python literals or strings.

    >>> from pyramid_heroku import expandvars_dict
    >>> from pprint import pprint
    >>> from unittest import mock
    >>> import os
    >>> with mock.patch.dict(os.environ,{'STRING':'text', 'BOOL': 'true', 'INT':'1', 'NESTED':'nested_${STRING}'}):
    ...    pprint(expandvars_dict({
    ...        "bool": "true",
    ...        "default_bool": "${FOO:-false}",
    ...        "default_int": "${FOO:-2}",
    ...        "default_string": "${FOO:-bar}",
    ...        "env_bool": "${BOOL:-foo}",
    ...        "env_int": "${INT:-foo}",
    ...        "env_string": "${STRING:-foo}",
    ...        "int": "1",
    ...        "must_be_set": "${STRING:? error: STRING must be set!}",
    ...        "nested": "${NESTED:-foo}",
    ...        "string": "foo",
    ...    }))
    {'bool': True,
     'default_bool': False,
     'default_int': 2,
     'default_string': 'bar',
     'env_bool': True,
     'env_int': 1,
     'env_string': 'text',
     'int': 1,
     'must_be_set': 'text',
     'nested': 'nested_text',
     'string': 'foo'}
    """
    return {k: safe_eval(expandvars(expandvars(v))) for k, v in settings.items()}



"""Tests for ClientAddr tween."""

from pyramid import request
from pyramid import testing
from zope.testing.loggingsupport import InstalledHandler

import logging
import mock
import responses
import structlog
import unittest

tweens_handler = InstalledHandler("pyramid_heroku.client_addr")


class TestClientAddrTween(unittest.TestCase):
    def wrap_logger(self, _, __, event_dict):
        """Have structlog output to the standard logger, so that tweens_handler can intercept its messages."""
        wrapped_logger = logging.getLogger("pyramid_heroku.client_addr")
        wrapped_logger.info(event_dict["event"])
        return event_dict["event"]

    def setUp(self):
        tweens_handler.clear()
        self.config = testing.setUp()
        self.request = request.Request({})

        self.handler = mock.Mock()
        self.registry = mock.Mock()
        self.registry.settings = {}

        self.responses = responses.RequestsMock()
        self.responses.start()
        self.responses.add(
            responses.GET,
            "https://www.cloudflare.com/ips-v4",
            status=200,
            body="8.8.8.0/24\n9.9.9.0/24",
        )
        self.responses.add(
            responses.GET,
            "https://www.cloudflare.com/ips-v6",
            status=200,
            body="2400:cb00::/32\n2606:4700::/32",
        )

        structlog.configure(processors=[self.wrap_logger], context_class=dict)

    def tearDown(self):
        self.responses.stop()
        self.responses.reset()
        tweens_handler.clear()
        testing.tearDown()

    def test_direct_access(self):
        from pyramid_heroku.client_addr import ClientAddr

        self.request.environ["REMOTE_ADDR"] = "1.2.3.4"
        ClientAddr(self.handler, self.registry)(self.request)
        self.handler.assert_called_with(self.request)
        self.assertEqual(self.request.client_addr, "1.2.3.4")

    def test_proxy_access(self):
        from pyramid_heroku.client_addr import ClientAddr

        self.request.environ["REMOTE_ADDR"] = "127.0.0.1"  # load balancer
        self.request.headers["X-Forwarded-For"] = "6.6.6.6,1.2.3.4"

        ClientAddr(self.handler, self.registry)(self.request)
        self.handler.assert_called_with(self.request)
        self.assertEqual(self.request.client_addr, "1.2.3.4")

    def test_spaces(self):
        from pyramid_heroku.client_addr import ClientAddr

        self.request.environ["REMOTE_ADDR"] = "127.0.0.1"  # load balancer
        self.request.headers["X-Forwarded-For"] = " 6.6.6.6, 1.2.3.4 "

        ClientAddr(self.handler, self.registry)(self.request)
        self.handler.assert_called_with(self.request)
        self.assertEqual(self.request.client_addr, "1.2.3.4")

    def test_cloudflare_ip_ignored(self):
        from pyramid_heroku.client_addr import ClientAddr

        self.request.environ["REMOTE_ADDR"] = "127.0.0.1"  # load balancer
        self.request.headers["X-Forwarded-For"] = "6.6.6.6, 1.2.3.4, 9.9.9.9"

        ClientAddr(self.handler, self.registry)(self.request)
        self.handler.assert_called_with(self.request)
        self.assertEqual(self.request.client_addr, "1.2.3.4")

    def test_cloudflare_ipv6_ignored(self):
        from pyramid_heroku.client_addr import ClientAddr

        self.request.environ["REMOTE_ADDR"] = "127.0.0.1"  # load balancer
        self.request.headers["X-Forwarded-For"] = (
            "2600:cb00:0000:0000:0000:0000:0000:0001, "
            "2500:cb00:0000:0000:0000:0000:0000:0001, "
            "2400:cb00:0000:0000:0000:0000:0000:0001"
        )

        ClientAddr(self.handler, self.registry)(self.request)
        self.handler.assert_called_with(self.request)
        self.assertEqual(
            self.request.client_addr, "2500:cb00:0000:0000:0000:0000:0000:0001"
        )

    def test_invalid_ips(self):
        from pyramid_heroku.client_addr import ClientAddr

        self.request.environ["REMOTE_ADDR"] = "127.0.0.1"  # load balancer
        self.request.headers["X-Forwarded-For"] = "1.2.3.4, invalid_ip not_ip"

        ClientAddr(self.handler, self.registry)(self.request)
        self.handler.assert_called_with(self.request)
        self.assertEqual(self.request.client_addr, "1.2.3.4")

    def test_cloudflare_ipv4_6_ignored(self):
        from pyramid_heroku.client_addr import ClientAddr

        self.request.environ["REMOTE_ADDR"] = "127.0.0.1"  # load balancer
        self.request.headers[
            "X-Forwarded-For"
        ] = "1.2.3.4, 9.9.9.9, 2400:cb00:0000:0000:0000:0000:0000:0001"

        ClientAddr(self.handler, self.registry)(self.request)
        self.handler.assert_called_with(self.request)
        self.assertEqual(self.request.client_addr, "1.2.3.4")

    def test_cloudflare_ip_list_get_error(self):
        from pyramid_heroku.client_addr import ClientAddr

        self.responses.reset()
        self.responses.add(
            responses.GET, "https://www.cloudflare.com/ips-v4", status=501, body="error"
        )
        self.responses.add(
            responses.GET, "https://www.cloudflare.com/ips-v6", status=501, body="error"
        )

        self.request.environ["REMOTE_ADDR"] = "127.0.0.1"  # load balancer
        self.request.headers["X-Forwarded-For"] = " 6.6.6.6, 1.2.3.4, 9.9.9.9"

        registry = mock.Mock()
        registry.settings = {}
        ClientAddr(self.handler, registry)(self.request)
        self.handler.assert_called_with(self.request)
        self.assertEqual(self.request.client_addr, "9.9.9.9")

        self.assertEqual(len(tweens_handler.records), 2)
        self.assertEqual(
            "Failed getting a list of IPs from https://www.cloudflare.com/ips-v4",
            tweens_handler.records[0].msg,
        )
        self.assertEqual(
            "Failed getting a list of IPs from https://www.cloudflare.com/ips-v6",
            tweens_handler.records[1].msg,
        )

        # structlog logging
        tweens_handler.clear()
        self.responses.reset()
        self.responses.add(
            responses.GET,
            "https://www.cloudflare.com/ips-v4",  # noqa
            status=501,
            body="error",
        )
        self.responses.add(
            responses.GET, "https://www.cloudflare.com/ips-v6", status=501, body="error"
        )

        registry.settings = {"pyramid_heroku.structlog": True}
        ClientAddr(self.handler, registry)(self.request)
        self.handler.assert_called_with(self.request)
        self.assertEqual(self.request.client_addr, "9.9.9.9")

        self.assertEqual(len(tweens_handler.records), 2)
        self.assertEqual(
            "Failed getting a list of IPs from https://www.cloudflare.com/ips-v4",
            tweens_handler.records[0].msg,
        )
        self.assertEqual(
            "Failed getting a list of IPs from https://www.cloudflare.com/ips-v6",
            tweens_handler.records[1].msg,
        )



"""Heroku API client."""

from requests import Response
from requests import Session
from typing import Optional

import os


class Heroku(object):

    api_endpoint = "https://api.heroku.com"

    def __init__(self, app_name: str, ini_file: str) -> None:
        """
        :param app_name: Name of Heroku app or id.
        :param ini_file: development.ini or production.ini filename.
        """
        self._formation = None
        self.app_name = app_name
        self.ini_file = ini_file

        headers = {
            "Authorization": f"Bearer {self.auth_key}",
            "Accept": "application/vnd.heroku+json; version=3",
            "Content-Type": "application/json",
        }
        self.session = Session()
        self.session.headers.update(headers)

    @property
    def auth_key(self) -> Optional[str]:
        """Heroku API secret.

        https://devcenter.heroku.com/articles/platform-api-quickstart#authentication.
        """
        return os.environ.get("MIGRATE_API_SECRET_HEROKU")

    def scale_down(self):
        """Scale all app workers to 0."""
        updates = [dict(type=t, quantity=0) for t in self.formation.keys()]
        res = self.session.patch(
            f"{self.api_endpoint}/apps/{self.app_name}/formation",
            json=dict(updates=updates),
        )
        self.parse_response(res)
        print("Scaled down to:")
        for x in res.json():
            print(f'{x["type"]}={x["quantity"]}')

    def scale_up(self):
        """Scale app back to initial state."""
        updates = [dict(type=t, quantity=s) for t, s in self.formation.items()]
        res = self.session.patch(
            f"{self.api_endpoint}/apps/{self.app_name}/formation",
            json=dict(updates=updates),
        )
        self.parse_response(res)
        print("Scaled up to:")
        for x in res.json():
            print(f'{x["type"]}={x["quantity"]}')

    @property
    def formation(self):
        """Get current app status and configuration.

        :return: Heroku app status as dict.
        """
        if not self._formation:
            res = self.session.get(
                f"{self.api_endpoint}/apps/{self.app_name}/formation"
            )
            self.parse_response(res)
            self._formation = {x["type"]: x["quantity"] for x in res.json()}
        return self._formation

    def get_maintenance(self) -> bool:
        res = self.session.get(f"{self.api_endpoint}/apps/{self.app_name}")
        self.parse_response(res)
        return res.json()["maintenance"]

    def set_maintenance(self, state: bool) -> None:
        res = self.session.patch(
            f"{self.api_endpoint}/apps/{self.app_name}", json=dict(maintenance=state)
        )
        if self.parse_response(res):
            print("Maintenance {}".format("enabled" if state else "disabled"))

    def parse_response(self, res: Response) -> Optional[bool]:
        """
        Parses Heroku API response.

        :param res: requests object
        :return: true if request succeeded
        """
        if res.status_code != 200:
            print(res.json())
        res.raise_for_status()
        return True



"""Set request.host from X-Forwarded-Host."""


def includeme(config):
    config.add_tween("pyramid_heroku.host.Host")


class Host(object):
    """Set request.host to proxied X-Forwarded-Host header.

    The tween checks if `X-Forwarded-Host` is present and if it is,
    overwrites the value of request.host with the value from the header.

    Note that this is a potential security risk in some environments and
    the `X-Forwarded-Host` header is not trusted by Heroku's Load Balancer:
    https://devcenter.heroku.com/articles/http-routing#heroku-headers

    The reason we need this is to support AutoIdle's Custom Domains:
    https://github.com/teamniteo/pyramid-realworld-example-app/issues/127#issuecomment-711894122
    """

    def __init__(self, handler, registry):
        self.handler = handler
        self.registry = registry

    def __call__(self, request):
        if request.headers.get("X-Forwarded-Host"):
            request.host = request.headers["X-Forwarded-Host"]
        return self.handler(request)



"""Tests for Heroku API client."""

from mock import call

import json
import mock
import responses
import unittest


class TestHeroku(unittest.TestCase):
    def setUp(self):
        from pyramid_heroku.heroku import Heroku

        self.Heroku = Heroku

    @mock.patch("pyramid_heroku.heroku.print")
    @responses.activate
    def test_default_formation(self, out):
        h = self.Heroku("test", "etc/production.ini")
        h._formation = {"type": "web", "quantity": 8999}
        responses.add(
            responses.PATCH,
            "https://api.heroku.com/apps/test/formation",  # noqa
            status=200,
            json=[{"type": "web", "quantity": 9001}],
        )
        h.scale_up()
        out.assert_has_calls([call("Scaled up to:"), call("web=9001")])

    @mock.patch("pyramid_heroku.heroku.print")
    @responses.activate
    def test_scale_down(self, out):
        h = self.Heroku("test", "etc/production.ini")
        responses.add(
            responses.GET,
            "https://api.heroku.com/apps/test/formation",  # noqa
            status=200,
            json=[{"type": "web", "quantity": 1}],
        )
        responses.add(
            responses.PATCH,
            "https://api.heroku.com/apps/test/formation",  # noqa
            status=200,
            json=[{"type": "web", "quantity": 0}],
        )
        h.scale_down()
        out.assert_has_calls([call("Scaled down to:"), call("web=0")])

    @mock.patch("pyramid_heroku.heroku.print")
    @responses.activate
    def test_scale_up(self, out):
        h = self.Heroku("test", "etc/production.ini")
        responses.add(
            responses.GET,
            "https://api.heroku.com/apps/test/formation",  # noqa
            status=200,
            json=[{"type": "web", "quantity": 1}],
        )
        responses.add(
            responses.PATCH,
            "https://api.heroku.com/apps/test/formation",  # noqa
            status=200,
            json=[{"type": "web", "quantity": 5}],
        )
        h.scale_up()
        out.assert_has_calls([call("Scaled up to:"), call("web=5")])

    @responses.activate
    def test_get_maintenance(self):
        h = self.Heroku("test", "etc/production.ini")
        responses.add(  # noqa
            responses.GET,
            "https://api.heroku.com/apps/test",
            status=200,
            body=json.dumps({"maintenance": True}),
        )
        assert h.get_maintenance()

    @mock.patch("pyramid_heroku.heroku.print")
    @responses.activate
    def test_maintenance_on(self, out):
        h = self.Heroku("test", "etc/production.ini")
        responses.add(
            responses.PATCH, "https://api.heroku.com/apps/test", status=200  # noqa
        )
        h.set_maintenance(True)
        out.assert_has_calls([call("Maintenance enabled")])

    @mock.patch("pyramid_heroku.heroku.print")
    @responses.activate
    def test_maintenance_off(self, out):
        h = self.Heroku("test", "etc/production.ini")
        responses.add(
            responses.PATCH, "https://api.heroku.com/apps/test", status=200  # noqa
        )
        h.set_maintenance(False)
        out.assert_has_calls([call("Maintenance disabled")])

    @mock.patch("pyramid_heroku.heroku.print")
    @mock.patch("pyramid_heroku.heroku.Heroku.parse_response", return_value=False)
    def test_set_maintanence_fail(self, pr, out):
        """Test that set_maintenance() doesn't print the maintenance state if the call to it failed."""
        h = self.Heroku("test", "etc/production.ini")
        h.set_maintenance(True)
        out.assert_not_called()



"""Tests for Heroku maintenance script."""


from pyramid_heroku.maintenance import main

import mock
import pytest
import responses
import unittest


class TestHerokuMaintenance(unittest.TestCase):
    def setUp(self):
        from pyramid_heroku.heroku import Heroku

        self.Heroku = Heroku

    def test_default_values(self):
        from pyramid_heroku.maintenance import main

        def assert_args(args, called_with):
            with mock.patch("sys.argv", ["migrate.py"] + args), mock.patch(
                "pyramid_heroku.maintenance.Heroku"
            ) as heroku:
                main()
            heroku.assert_called_with(*called_with)

        # all default
        assert_args(["on", "my_test_app"], ["my_test_app", "etc/production.ini"])

        # all custom
        assert_args(
            ["on", "my_test_app", "etc/develop2.ini"],
            ["my_test_app", "etc/develop2.ini"],
        )

    @mock.patch("pyramid_heroku.maintenance.Heroku")
    @mock.patch("pyramid_heroku.heroku.Session")
    @responses.activate
    def test_migrate_on(self, sess, Heroku):

        with mock.patch("sys.argv", ["maintenance.py", "on", "my_app"]):
            main()

        Heroku().set_maintenance.assert_called_once_with(True)

    @mock.patch("pyramid_heroku.maintenance.Heroku")
    @mock.patch("pyramid_heroku.heroku.Session")
    @responses.activate
    def test_migrate_off(self, sess, Heroku):

        with mock.patch("sys.argv", ["maintenance.py", "off", "my_app"]):
            main()

        Heroku().set_maintenance.assert_called_once_with(False)

    @mock.patch("pyramid_heroku.subprocess")
    @mock.patch("pyramid_heroku.print")
    @mock.patch("pyramid_heroku.heroku.Session")
    @responses.activate
    def test_maintenance_non_zero(self, ses, out, sub):
        h = self.Heroku("test", "etc/production.ini")
        p = mock.Mock()
        p.stdout = b"foo"
        p.stderr = b"bar"
        p.check_returncode.side_effect = Exception
        sub.run.return_value = p

        with pytest.raises(Exception):
            main(h)

        assert not out.called



"""Tests for utils."""

import pytest


def test_safe_eval():
    """Test that dangerous values are safely eval-ed."""
    from pyramid_heroku import safe_eval

    assert safe_eval("False") is False
    assert safe_eval("True") is True
    assert safe_eval("false") is False
    assert safe_eval("true") is True
    assert safe_eval("1337") == 1337
    assert safe_eval("1337.1337") == 1337.1337
    assert safe_eval("asdfghjkl") == "asdfghjkl"
    assert safe_eval("asdfg hjkl") == "asdfg hjkl"
    assert safe_eval("'asdfg hjkl'") == "asdfg hjkl"
    assert safe_eval("u'true'") == "true"
    assert (
        safe_eval("raise Exception('Chrashed yet?')")
        == "raise Exception('Chrashed yet?')"
    )

    assert safe_eval("exit") == "exit"
    assert safe_eval("exit(12)") == "exit(12)"
    assert safe_eval("None") is None
    assert safe_eval("('test', 'tIsT')") == ("test", "tIsT")
    assert safe_eval("(('fOo', 'BaRrAr'), ('asd', 'AsD'))") == (
        ("fOo", "BaRrAr"),
        ("asd", "AsD"),
    )

    assert safe_eval("") == ""
    assert safe_eval("[]") == []
    with pytest.raises(ValueError):
        assert safe_eval(None) == []
    with pytest.raises(ValueError):
        assert safe_eval([1, 2, 3]) == []


def test_expandvars_dict():
    """Test that env vars are added to settings dict."""
    from pyramid_heroku import expandvars_dict
    import os

    os.environ["FOO"] = "foo"
    os.environ["BAR"] = "bar"
    os.environ["NESTED"] = "${FOO}bar"
    settings = {
        "test_boolean": "true",
        "test_integer": "1337",
        "test_string": "'asdfg hjkl'",
        "test_none": "None",
        "test_set": "('test', 'tIsT')",
        "test_multi_set": "(('fOo', 'BaRrAr'), ('asd', 'AsD'))",
        "test_empty": "",
        "test_env": "${FOO}",
        "test_env_nested": "${NESTED}",
        "test_dollar": "$",
        "test_endswith_dollar": "BAR$",
        "test_simple_variable": "$FOO",
        "test_multi_variables": "$FOO${BAR}",
        "test_default_set": "${FOO:-default}",
        "test_default_unset": "${BUZ:-default}",
        "test_substitute_set": "${FOO:+default}",
        "test_substitute_unset": "${BUZ:+default}",
        "test_offset": "${BAR:2}",
        "test_offset_length": "${BAR:1:3}",
    }

    expanded_settings = {
        "test_boolean": True,
        "test_integer": 1337,
        "test_string": "asdfg hjkl",
        "test_none": None,
        "test_set": ("test", "tIsT"),
        "test_multi_set": (("fOo", "BaRrAr"), ("asd", "AsD")),
        "test_empty": "",
        "test_env": "foo",
        "test_env_nested": "foobar",
        "test_dollar": "$",
        "test_endswith_dollar": "BAR$",
        "test_simple_variable": "foo",
        "test_multi_variables": "foobar",
        "test_default_set": "foo",
        "test_default_unset": "default",
        "test_substitute_set": "default",
        "test_substitute_unset": "",
        "test_offset": "r",
        "test_offset_length": "ar",
    }

    assert expanded_settings == expandvars_dict(settings)
